
#define __NO_VERSION__

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/wait.h>
#include <linux/isdnif.h>
#include <linux/tty.h>


#include "ih_card.h"

#define HSCX_TEST_IRQ

static byte readisac_pcc(struct isachscx *ih, byte offset)
{
	byte val;
	long flags;
	save_flags(flags);
	cli();
	outb_p(offset,ih->iobase+3);
	val = inb_p(ih->iobase);
	restore_flags(flags);
	return val;
}

static byte readisac_0(struct isachscx *ih, byte offset)
{
	return readb(ih->membase + ((offset & 1) ? 0x2ff : 0x100) + offset);
}

static byte readisac_3(struct isachscx *ih, byte offset)
{
	return (inb(ih->iobase - 0x420 + offset));
}

static void writeisac_0(struct isachscx *ih, byte offset, byte value)
{
	writeb(value,ih->membase + ((offset & 1) ? 0x2ff : 0x100) + offset);
}

static void writeisac_pcc(struct isachscx *ih, byte offset, byte value)
{
	long flags;
	save_flags(flags);
	cli();
	outb_p(offset,ih->iobase+3);
	outb_p(value,ih->iobase);
	restore_flags(flags);
}

static void writeisac_3(struct isachscx *ih, byte offset, byte value)
{
	outb(value,ih->iobase - 0x420 + offset);
}

static void readisac_s_0(struct isachscx *ih,byte * dest, int count)
{
	while (count--)
		*dest++ = readisac_0(ih,0x0);
}

static void readisac_s_3(struct isachscx *ih,byte * dest, int count)
{
	insb(ih->iobase - 0x420 + 0x3e, dest, count);
}

static void readisac_s_pcc(struct isachscx *ih,byte *dest,int cnt)
{
	long flags;

	save_flags(flags);
	cli();
	outb(0x0,ih->iobase+3);
	insb(ih->iobase,dest,cnt);
	restore_flags(flags);
}

static void writeisac_s_0(struct isachscx *ih,byte *src,int cnt)
{
	while (cnt--)
		writeisac_0(ih, 0x0, *src++);
}
 
static void writeisac_s_pcc(struct isachscx *ih,byte *src,int cnt)
{
	long flags;

	save_flags(flags);
	cli();
	outb(0x0,ih->iobase+3);
	outsb(ih->iobase,src,cnt);
	restore_flags(flags);
}

static void writeisac_s_3(struct isachscx *ih,byte * src, int count)
{
	outsb(ih->iobase - 0x420 + 0x3e, src, count);
}

static byte readhscx_0(struct isachscx *ih, byte hscx, byte offset)
{
	return readb(ih->membase + ((offset & 1) ? 0x1FF : 0) + ((hscx & 1) ? 0x1c0 : 0x180) + offset);
}

static byte readhscx_pcc(struct isachscx *ih, byte hscx, byte offset)
{
	byte val;
	long flags;
	save_flags(flags);
	cli();
	outb(offset+(hscx<<6),ih->iobase+3);
	val = inb(ih->iobase+2);
	restore_flags(flags);
	return val;
}

static byte readhscx_3(struct isachscx *ih, byte hscx, byte offset)
{
	return (inb(ih->iobase - (hscx ? 0x820 : 0xc20) + offset));
}

static void writehscx_0(struct isachscx *ih, byte hscx, byte offset, byte data)
{
	writeb(data,ih->membase + ((offset & 1) ? 0x1FF : 0) + ((hscx & 1) ? 0x1c0 : 0x180) + offset);
}

static void writehscx_pcc(struct isachscx *ih, byte hscx, byte offset, byte data)
{
	long flags;
	save_flags(flags);
	cli();
	outb(offset+(hscx<<6),ih->iobase+3);
	outb(data,ih->iobase+2);
	restore_flags(flags);
}

static void writehscx_s_pcc(struct isachscx *ih, byte hscx, byte * src , int cnt)
{
	long flags;

	save_flags(flags);
	cli();
	outb(0x0+(hscx<<6),ih->iobase+3);
	outsb(ih->iobase+2,src,cnt);
	restore_flags(flags);
}

static void writehscx_s_0(struct isachscx *ih, byte hscx,byte * src , int cnt)
{
	while (cnt--)
		writehscx_0(ih,hscx, 0x0, *src++);
}

static void readhscx_s_0(struct isachscx *ih, byte hscx,byte *dest, int cnt)
{
	while (cnt--)
		*dest++ = readhscx_0(ih, hscx,0x0);
}

static void readhscx_s_pcc(struct isachscx *ih, byte hscx,byte *dest, int cnt)
{
	long flags;

	save_flags(flags);
	cli();
	outb(0x0+(hscx<<6),ih->iobase+3);
	insb(ih->iobase+2,dest,cnt);
	restore_flags(flags);
}

static void writehscx_3(struct isachscx *ih, byte hscx, byte offset, byte data)
{
	outb(data,ih->iobase - (hscx ? 0x820 : 0xc20) + offset);
}

static void readhscx_s_3(struct isachscx *ih, byte hscx,byte * dest, int count)
{
	insb(ih->iobase - (hscx ? 0x820 : 0xc20) + 0x3e, dest, count);
}

static void writehscx_s_3(struct isachscx *ih, byte hscx,byte * src, int count)
{
	outsb(ih->iobase - (hscx ? 0x820 : 0xc20) + 0x3e, src, count);
}

/************************/

inline void writehscx(struct isachscx *ih, byte hscx, byte offset,byte data)
{
	(ih->writehscx)(ih,hscx,offset,data);
}

inline void readhscx_s(struct isachscx *ih, byte hscx, byte * dest, int count)
{
	(ih->readhscx_s)(ih,hscx,dest,count);
}

inline void writehscx_s(struct isachscx *ih, byte hscx,byte * src, int count)
{
	(ih->writehscx_s)(ih,hscx,src,count);
}

inline byte readhscx(struct isachscx *ih, byte hscx, byte offset)
{
	return (ih->readhscx)(ih,hscx,offset);
}

inline void writeisac(struct isachscx *ih, byte offset, byte value)
{
	(ih->writeisac)(ih,offset,value);
}

inline byte readisac(struct isachscx *ih, byte offset)
{
	return (ih->readisac)(ih,offset);
}

inline void readisac_s(struct isachscx *ih,byte *dest,int cnt)
{
	(ih->readisac_s)(ih,dest,cnt);
}

inline void writeisac_s(struct isachscx *ih,byte *src,int cnt)
{
	(ih->writeisac_s)(ih,src,cnt);
}

void writehscxCMDR(struct isachscx *ih, int hscx, unsigned char data)
{
	long            flags,to=10;

	save_flags(flags);
	cli();
	while ((readhscx(ih, hscx, HSCX_STAR) & 0x04)) {
		udelay(5);
		to--;
		if (to<=0) {
			printk(KERN_WARNING "HSCX CMD timeout\n");
			return;
		}
	}
	writehscx(ih, hscx, HSCX_CMDR, data);
	restore_flags(flags);
}

void ih_initisac(struct isachscx *ih,int iom1_adf1,int iom2_adf1)
{
	writeisac(ih, ISAC_MASK, 0xff);

	switch(ih->cardtype) {
		case TYPE_16_0:
			/* IOM-1 */
			writeisac(ih, ISAC_ADF2, 0x0);
			writeisac(ih, ISAC_SPCR, 0xa);
			writeisac(ih, ISAC_ADF1, iom1_adf1);
			break;
		case TYPE_16_3:
		case TYPE_PCC:
			/* IOM-2 */
			writeisac(ih, ISAC_ADF2, 0x80);
			writeisac(ih, ISAC_SQXR, 0x2f);
			writeisac(ih, ISAC_SPCR, 0x00);
			writeisac(ih, ISAC_TIMR, 0x00);
			writeisac(ih, ISAC_ADF1, iom2_adf1);
			break;
	}
	writeisac(ih, ISAC_STCR, 0x70);
	writeisac(ih, ISAC_MODE, 0xc9);
	writeisac(ih, ISAC_CMDR, 0x41);
	writeisac(ih, ISAC_CIX0, (1 << 2) | 3);
}

static int probe_hscx(struct isachscx *ih)
{
	unsigned char valA,valB;
	char *name;

	if(ih->name)
		name = ih->name;
	else
		name = "ISDN Card";

	valA = readhscx(ih,0,HSCX_VSTR) & 0x7f;
	valB = readhscx(ih,1,HSCX_VSTR) & 0x7f;
	if(valA == valB && valA >= 0x4 && valA < 0xf) {
		writehscx(ih,0,HSCX_CCR1,0x85);
		writehscx(ih,0,HSCX_CCR2,0x30);
		valA = readhscx(ih,0,HSCX_CCR1);
		valB = readhscx(ih,0,HSCX_CCR2);
		if(valA == 0x85 && valB == 0x30) {
#ifdef HSCX_TEST_IRQ
			int irqval;

			autoirq_setup(2);
			ih_hscxmode(ih,0, 2, 0,7);
			irqval = autoirq_report(2);
			if(!irqval && ih->irq <=0) {
				ih_hscxmode(ih,0, 0, 0,7);
				printk(KERN_ERR "%s: Can't get interrupt: %d\n",name,irqval);
				return 0;
			}
			if(ih->irq > 0 && ih->irq != irqval) {
				printk(KERN_WARNING "%s: Can't verify interrupt: wanted %d, got %d\n",
					name,ih->irq,irqval);
			}
			if(ih->irq <=0)
				ih->irq = irqval;
			ih_hscxmode(ih,0, 0, 0,7);
#else
			writehscx(ih,0,HSCX_CCR1,0x05);
#endif
			return 1;
		}
		printk(KERN_INFO "%s: can't set HSCX CCR1 register: %d \n",name,valA);
	}
	else 
		printk(KERN_INFO "%s: unknown HSCX versions: A:%d B:%d\n",name,valA,valB);
	return 0;
}

/* 
 * ELSA PCC 16 card probe and timer check functions
 */
static int probe_pcc_timer(struct isachscx *ih)
{
	int i,addr=ih->iobase;

	inb(addr+6);

	i = 0;
	while (!(inb(addr+5)&0x2)) {
		i++;
		udelay(100);
		if(i>1000) {
			printk(KERN_ERR "ELSA ISDN/PCC-16: timer doesn't start!\n");
			return 0;
		}
	}
	i = 0;
	while ((inb(addr+5)&0x2)) {
		i++;
		udelay(100);
		if(i>1000) {
			printk(KERN_ERR "ELSA ISDN/PCC-16: no timer timeout!\n");
			return 0;
		}
	}
	return 1;
}

static int probe_pcc_16(struct isachscx *ih)
{
	unsigned char version,cfg;
	int i,irqval,irqsave,addr = ih->iobase;
	static int val2irq[] = { -1 , -1 , 15 , 11 , 10 , 5 , 3 , 2 };
	static int iovals[] = { 0x160,0x170,0x260,0x360 };
	static char name[] = "ELSA PCC-16";

	if(check_region(ih->iobase, 8))
		return 0;

	for(i=0;i<4;i++)
		if(ih->iobase == iovals[i])
			break;
	if(i == 4)
		return 0;
	
	cfg = inb_p(addr+5);
	for(i=0;i<8;i++) { /* toggle a few times */
		cfg ^= 0x4;
		if( (cfg & 0x4) != (inb_p(addr+5) & 0x4) ) {
			printk(KERN_INFO "%s: Can't find card at %#x!\n",name,addr);
			return 0;
		}
	}

	ih->name = name;

	version = inb_p(addr+3);
	irqval = (cfg>>3)&0x7;

	if(val2irq[irqval] < 0) {
		printk(KERN_INFO "%s: strange irqval in CFG register\n",name);
		return 0;
	}

	if(!probe_pcc_timer(ih))
		return 0;
	outb(0x00,addr+4);	/* reset */
	if(!probe_pcc_timer(ih))
		return 0;
	outb(0x20,addr+4);	/* unreset */
	if(!probe_pcc_timer(ih))
		return 0;

	ih->pcc_ctrl = 0x20;

#ifdef USE_PCC_TIMER
	ih->pcc_ctrl |= 0x80;
	outb(ih->pcc_ctrl,addr+4);
	inb(addr+6);	/* start timer */
#endif
	ih->cardtype = TYPE_PCC;
	ih->writehscx = writehscx_pcc;
	ih->readhscx = readhscx_pcc;
	ih->writehscx_s = writehscx_s_pcc;
	ih->readhscx_s = readhscx_s_pcc;
	ih->writeisac = writeisac_pcc;
	ih->readisac = readisac_pcc;
	ih->writeisac_s = writeisac_s_pcc;
	ih->readisac_s = readisac_s_pcc;

	irqsave = ih->irq;
	ih->irq = val2irq[irqval];

	if(!probe_hscx(ih)) {
		ih->irq = irqsave;
		printk(KERN_INFO "%s: Can't find ELSA card. (no HSCX)\n",name);
		return 0;
	}

	request_region(addr,8,name);
	return 1;
}
	
static int probe_16_3(struct isachscx *ih) 
{
	static char name[] = "Teles 16.3";
	int addr_save = ih->iobase;
	static signed char irq2val[] = { -1,-1,0x00,0x02,0x04,0x06,-1,-1,-1,-1,0x08,0x0a,0x0c,-1,-1,0x0e };
	unsigned long timout;
	byte cfval;

	if(!ih->iobase || ih->membase)
		return 0;

	if(ih->irq < 0 || ih->irq > 15 || irq2val[ih->irq] < 0)
		cfval = 0x00;
	else
		cfval = irq2val[ih->irq];

	ih->iobase = (ih->iobase & 0x3ff) + 0xc00;

	if(check_region(ih->iobase-0x000, 32) ||
	  check_region(ih->iobase-0x400, 32) || 
	  check_region(ih->iobase-0x800, 32) ||
	  check_region(ih->iobase-0xc00, 32)) {
		ih->iobase = addr_save;
		return 0;
	}

	ih->cardtype = TYPE_16_3;
	ih->name = name;
	ih->writehscx = writehscx_3;
	ih->readhscx = readhscx_3;
	ih->writehscx_s = writehscx_s_3;
	ih->readhscx_s = readhscx_s_3;
	ih->writeisac = writeisac_3;
	ih->readisac = readisac_3;
	ih->writeisac_s = writeisac_s_3;
	ih->readisac_s = readisac_s_3;

	timout = jiffies + (HZ / 10) + 1;
	outb(cfval,ih->iobase + 4);
	while (jiffies <= timout);

	timout = jiffies + (HZ / 10) + 1;
	outb(cfval | 1,ih->iobase + 4);
	while (jiffies <= timout);

	if(!probe_hscx(ih)) {
		ih->iobase = addr_save;
		return 0;
	}

	request_region(ih->iobase-0x000, 32, name);
	request_region(ih->iobase-0x400, 32, name);
	request_region(ih->iobase-0x800, 32, name);
	request_region(ih->iobase-0xc00, 32, name);

	return 1;
}


static int probe_16_0(struct isachscx *ih)
{
	static signed char irq2val[] = { -1,-1,0x00,0x02,0x04,0x06,-1,-1,-1,-1,0x08,0x0a,0x0c,-1,-1,0x0e };
	static char name[] = "Teles 16.0";
	unsigned long timout;
	byte            cfval, val;

	if(!ih->iobase && !ih->membase)
		return 0;

	if (check_region(ih->iobase, 8)) {
		return 0;
	}

	if(ih->irq < 0 || ih->irq > 15 || irq2val[ih->irq] < 0)
		cfval = 0x00;
	else
		cfval = irq2val[ih->irq];

	cfval |= (((unsigned int) ih->membase >> 9) & 0xF0);

	if (inb(ih->iobase + 0) != 0x51) {
		printk(KERN_INFO "XXX Byte at %x is %x\n",
		    ih->iobase + 0, inb(ih->iobase + 0));
		return 0;
	}
	if (inb(ih->iobase + 1) != 0x93) {
		printk(KERN_INFO "XXX Byte at %x is %x\n",
		   ih->iobase + 1, inb(ih->iobase + 1));
		return 0;
	}
	val = inb(ih->iobase + 2); /* 0x1e=without AB
					* 0x1f=with AB
					*/
	if (val != 0x1e && val != 0x1f) {
		printk(KERN_INFO "XXX Byte at %x is %x\n",
		       ih->iobase + 2, inb(ih->iobase + 2));
		return 0;
	}
	request_region(ih->iobase, 8, name);

	ih->name = name;
	ih->cardtype = TYPE_16_0;
	ih->writehscx = writehscx_0;
	ih->readhscx = readhscx_0;
	ih->writehscx_s = writehscx_s_0;
	ih->readhscx_s = readhscx_s_0;
	ih->writeisac = writeisac_0;
	ih->readisac = readisac_0;
	ih->writeisac_s = writeisac_s_0;
	ih->readisac_s = readisac_s_0;

	timout = jiffies + (HZ / 10) + 1;
	outb(cfval,ih->iobase + 4);
	while (jiffies <= timout);

	timout = jiffies + (HZ / 10) + 1;
	outb(cfval | 1,ih->iobase + 4);
	while (jiffies <= timout);

	timout = jiffies + (HZ / 5) + 1;
	writeb(0,ih->membase + 0x80);
	while (jiffies <= timout);

	writeb(1,ih->membase + 0x80);
	timout = jiffies + (HZ / 5) + 1;
	while (jiffies <= timout);

	return 1;
}

static int probe_8(struct isachscx *ih)
{
	static char name[] = "Teles 8";
	unsigned long timout;

	if(ih->iobase || !ih->membase)
		return 0;

	ih->cardtype = TYPE_16_0; /*same access scheme for 16_0 and 8 */
	ih->name = name;
	ih->writehscx = writehscx_0;
	ih->readhscx = readhscx_0;
	ih->writehscx_s = writehscx_s_0;
	ih->readhscx_s = readhscx_s_0;
	ih->writeisac = writeisac_0;
	ih->readisac = readisac_0;
	ih->writeisac_s = writeisac_s_0;
	ih->readisac_s = readisac_s_0;

	timout = jiffies + (HZ / 5) + 1;
	writeb(0,ih->membase + 0x80);
	while (jiffies <= timout);
	writeb(1,ih->membase + 0x80);
	timout = jiffies + (HZ / 5) + 1;
	while (jiffies <= timout);

	return 1;
}


int ih_checkcard(struct isachscx *ih,int iobase,void *membase,int irq)
{
	char *proto = "UNKNOWN";

	ih->cardtype = TYPE_UNKNOWN;
	ih->name = "ISDN Card";
	ih->iobase = iobase;
	ih->membase = membase;
	ih->irq = irq;

	if(probe_8(ih)) {
		printk(KERN_NOTICE "%s assumed, mem: %lx irq: %d proto: %s\n",ih->name,
			(long) ih->membase, ih->irq,proto);
	}
	else if(probe_pcc_16(ih)) {
		printk(KERN_NOTICE "%s found, io: %#x, irq: %d, proto: %s\n",
			ih->name,ih->iobase,ih->irq,proto);
	}
	else if(probe_16_3(ih)) {
		printk(KERN_NOTICE "%s found, io: %x irq: %d proto: %s\n",
			ih->name,ih->iobase, ih->irq,proto);
	}
	else if(probe_16_0(ih)) {
		printk(KERN_NOTICE "%s found, io: %x mem: %lx irq: %d proto: %s\n",
			ih->name,ih->iobase, (long) ih->membase, ih->irq,proto);
	}
	else 
		return -1;

	printk(KERN_INFO "HSCX version A:%x B:%x\n",
	      readhscx(ih, 0, HSCX_VSTR) & 0xf,
	      readhscx(ih, 1, HSCX_VSTR) & 0xf);

	return (0);
}

void ih_hscxmode(struct isachscx *ih,int hscx, int mode, int ichan,int chanbits)
{
	static unsigned char iom1slot[3] = { 0x07 , 0x03 };
	static unsigned char iom2slot[3] = { 0x2f , 0x03 };
	static unsigned char modes[] = { -1 ,  0xe4 , 0x8c , 0xe5 , 0x8d } ;

	if(mode == 0) {
		writehscx(ih, hscx, HSCX_MASK, 0xff);
		writehscx(ih, hscx, HSCX_MODE, 0x80);
		writehscx(ih, hscx, HSCX_CCR1, 0x05);        /* power down */
		return;
	}

	writehscx(ih, hscx, HSCX_CCR1, 0x85);
	writehscx(ih, hscx, HSCX_XAD1, 0xFF);
	writehscx(ih, hscx, HSCX_XAD2, 0xFF);
	writehscx(ih, hscx, HSCX_RAH2, 0xFF);
	writehscx(ih, hscx, HSCX_XBCH, 0x00);
	writehscx(ih, hscx, HSCX_RLCR, 0x00);
	writehscx(ih, hscx, HSCX_CCR2, 0x30);

	switch(ih->cardtype) {
		case TYPE_16_0:
			writehscx(ih, hscx, HSCX_TSAX, iom1slot[ichan]);
			writehscx(ih, hscx, HSCX_TSAR, iom1slot[ichan]);
			break;
		case TYPE_16_3:
		case TYPE_PCC:
			writehscx(ih, hscx, HSCX_TSAX, iom2slot[ichan]);
			writehscx(ih, hscx, HSCX_TSAR, iom2slot[ichan]);
			break;
	}

	writehscx(ih, hscx, HSCX_XCCR, chanbits);
	writehscx(ih, hscx, HSCX_RCCR, chanbits);
	writehscx(ih, hscx, HSCX_MODE, modes[mode] );
	writehscxCMDR(ih, hscx, 0x41);
	writehscx(ih, hscx, HSCX_MASK, 0x00);
}

